-- HUD STUFF

-- Various graphics.
local tags = {
	neutral = {},
	angry = {}
}
local clawgfx = nil
local generatedtags = false

-- Cache fonts!
WTXT.createNewFont({
	name = "FANCYMENU",
	prefix = "MNUFN",
	monowidth = 8,
	xsepwidth = -2, -- The menu font is actually a bit wide...
})
local cachedfonts = false

-- Displays the current partner tag.
local function partnertag(v, p, c)
	if not (combinator.active) then return end
	
	if not (p.mo and p.mo.valid) then return end
	local mo = p.mo
	
	local partner, gargoyle = combinator.returnPartnerMobj(p, true)
	local cm = p.combiVars
	if not cm then return end
	if partner == nil then return end
	--if splitscreen >= 1 then return end
	
	local poszoffset = ((partner.height >> 1) + (36*partner.scale))*P_MobjFlip(partner)
	if gargoyle then poszoffset = partner.height*P_MobjFlip(partner) end
	
	local result = SG_ObjectTracking(v, p, c, {x = partner.x, y = partner.y, z = partner.z + poszoffset}, false)
	local basex, basey = result.x, result.y
	
	if not generatedtags then
		-- neutral tags
		tags.neutral = {[0] = v.cachePatch("CB_PARTA0"), [1] = v.cachePatch("CB_PARTB0")}
		generatedtags = true
	end
	
	local clr = v.getColormap(TC_RAINBOW, (partner.player and partner.player.skincolor or partner.color))
	local camDist = R_PointToDist2(c.x, c.y, partner.x, partner.y)
	if not result.onScreen then return end
	if camDist >= 2048*mapobjectscale then return end
	if not P_CheckSight(mo, partner) then return end
	
	if v.interpolate ~= nil then v.interpolate(true) end
	
	-- There is no easter bunny, there is no tooth fairy, and there is no r_splitscreen
	local offx, offy = SG_GetSplitscreenOffsets(v, c.pnum - 1)
	-- do a funny clip rect!
	if v.setClipRect ~= nil then
		local screenWidth, screenHeight = v.width(), v.height()
		local screenScaleX, _frac = v.dupx()
		local screenScaleY, _frac = v.dupy()
		local gameWidth, gameHeight = screenWidth / screenScaleX, screenHeight / screenScaleY
		local r_splitscreen = SG_GetSplitscreen()
		local cnum = c.pnum - 1
		
		if r_splitscreen == 1 then
			v.setClipRect(
				0, 
				((cnum == 1) and (gameHeight/2)*FRACUNIT or 0),
				gameWidth*FRACUNIT,
				(gameHeight/2)*FRACUNIT,
				V_SNAPTOLEFT|V_SNAPTOTOP
			)
		elseif r_splitscreen > 1 then
			v.setClipRect(
				((cnum & 1) and (gameWidth/2)*FRACUNIT or 0),
				((cnum > 1) and (gameHeight/2)*FRACUNIT or 0),
				(gameWidth/2)*FRACUNIT,
				(gameHeight/2)*FRACUNIT,
				V_SNAPTOLEFT|V_SNAPTOTOP
			)
		end
	end
	v.drawScaled(basex + offx, basey + offy, FRACUNIT, tags.neutral[((leveltime / 7) & 1)], ((P_MobjFlip(partner) == -1) and V_VFLIP or 0), clr)
	
	if v.interpolate ~= nil then v.interpolate(false) end
	
	if v.clearClipRect ~= nil then v.clearClipRect() end
end

-- for the text at the start
local function V_DanceYOffset(counter)
    local duration = 16
    local step = (leveltime + counter) % duration

    return abs(step - (duration / 2)) - (duration / 4)
end

-- for le claw
local finalclawpos = -24
local clawduration = TICRATE

-- peace, peace
local function K_DrawSticker(v, x, y, width, flags, small)
	local stickerend, height = nil, nil
	if small then
		stickerend = v.cachePatch("K_STIKE2")
		height = 6
	else
		stickerend = v.cachePatch("K_STIKEN")
		height = 11
	end
	
	v.draw(x, y, stickerend, flags)
	v.drawFill(x, y, width, height, 24|flags)
	v.draw(x+width, y, stickerend, flags|V_FLIP)
end

-- Redone hud code because fuck splitscreen
local function newhudinfostuff(v, p, c)
	if not (combinator.active) then return end
	if p.spectator then return end
	-- if splitscreen then return end
	
	if not (p.mo and p.mo.valid) then return end
	local mo = p.mo
	
	local partner, gargoyle = combinator.returnPartnerMobj(p, true)
	local cm = p.combiVars
	if not cm then return end
	
	-- force cache
	if not cachedfonts then WTXT.fontSentinel(v) cachedfonts = true end
	
	-- claw gfx
	if not clawgfx then
		clawgfx = {v.cachePatch("CB_CLWA0"), v.cachePatch("CB_CLWB0"), v.cachePatch("CB_CSWA0"), v.cachePatch("CB_CSWB0"), v.cachePatch("CB_CMWA0")}
	end
	
	-- get splitscreen state
	local r_splitscreen = SG_GetSplitscreen()
	local cnum = c.pnum - 1
	
	-- graphics and positions
	local clawclose = (r_splitscreen == 0 and clawgfx[2] or (r_splitscreen > 1 and clawgfx[5] or clawgfx[4]))
	local clawopen = (r_splitscreen == 0 and clawgfx[1] or (r_splitscreen > 1 and clawgfx[5] or clawgfx[3]))
	local clawiconidx, clawgidx = ((r_splitscreen > 1) and C or A), ((r_splitscreen > 1) and 1 or 0)
	local clawicon = ((gargoyle or partner == nil) and v.cachePatch(string.format("CB_GGLE%d", clawgidx)) or v.getSprite2Patch(partner.skin, SPR2_XTRA, false, clawiconidx))
	local clawiconhwidth = clawicon.width/2
	local clawiconcolor = (partner and v.getColormap(partner.skin, partner.color) or nil)
	if partner ~= nil and cm.fuckedOverTime and (cm.fuckedOverTime & 1) then
		clawiconcolor = v.getColormap(TC_ALLWHITE, partner.color)
	end
	local clawplayer = (gargoyle and "anti-loneliness gargoyle" or partner.player.name)
	
	-- so that i dont have to edit each one by one
	local flags = V_SNAPTOLEFT|V_SNAPTOTOP
	
	-- give it some hud transparency
	if r_splitscreen == 1 then
		flags = $ | V_30TRANS
	end
	
	-- grab true screen width
	-- snap flags are a fuck
	-- i do not want to see them again
	-- fuck splitscreen too
	local screenWidth, screenHeight = v.width(), v.height()
	local screenScaleX, _frac = v.dupx()
	local screenScaleY, _frac = v.dupy()
	local gameWidth, gameHeight = screenWidth / screenScaleX, screenHeight / screenScaleY
	
	-- claw positions
	local iconpos = ((r_splitscreen > 1) and 19 or 24)
	-- splitscreen bullshit
	local clawx, clawy = gameWidth / 2, 0
	local clawx_f, clawy_f = clawx*FRACUNIT, clawy*FRACUNIT
	-- x coords
	if (r_splitscreen > 1) then
		if cnum == 0 or cnum == 2 then
			clawx = $ - (gameWidth/4)
		else
			clawx = $ + (gameWidth/4)
		end
	end
	-- y coords, just move it down...
	if (r_splitscreen == 1 and (cnum == 1)) or (r_splitscreen > 1 and (cnum > 1)) then
		clawy = $ + (gameHeight/2)
	end
	
	-- update these ofc
	clawx_f, clawy_f = clawx*FRACUNIT, clawy*FRACUNIT
	local stupidoffset = ((r_splitscreen > 1) and -FRACUNIT or -FRACUNIT/2) -- yeah
	
	-- Show the claw.
	-- Claw animations are not shown on splitscreen...
	-- Wish I could, though you can't use v.drawCropped because it doesnt FUCKING EXIST
	--
	-- update: mr made to expose v_setcliprect and co
	
	if v.interpolate ~= nil then v.interpolate(true) end
	
	if (v.setClipRect == nil) then
		if (r_splitscreen > 0) and cm.finishstart == -1 then
			-- Normal drawing here...
			v.draw(clawx, clawy, clawclose, flags)
			-- Yes the offset is very important
			if partner ~= nil then
				v.drawScaled(clawx_f - ((clawiconhwidth*FRACUNIT) + stupidoffset), clawy_f + iconpos*FRACUNIT, FRACUNIT, clawicon, flags, clawiconcolor)
				v.drawString(clawx, clawy + iconpos + clawiconhwidth*2, clawplayer, flags | V_TRANSLUCENT, "thin-center")
			else
				local idx = ((leveltime / 3) % table.maxn(combinator.playerList)) + 1
				local mmap = v.getSprite2Patch(combinator.playerList[idx].mo.skin, SPR2_XTRA, false, clawiconidx)
				local clr = v.getColormap(combinator.playerList[idx].mo.skin, combinator.playerList[idx].mo.color)
				v.drawScaled(clawx_f - ((clawiconhwidth*FRACUNIT) + stupidoffset), clawy_f + iconpos*FRACUNIT, FRACUNIT, mmap, flags, clr)
			end
			return
		end
	
		if (r_splitscreen > 0) and cm.finishstart ~= -1 then return end -- just remove the damn thing
	else -- set clip rect, if its exposed
		if r_splitscreen == 1 then
			v.setClipRect(
				0, 
				((cnum == 1) and (gameHeight/2)*FRACUNIT or 0),
				gameWidth*FRACUNIT,
				(gameHeight/2)*FRACUNIT,
				flags
			)
		elseif r_splitscreen > 1 then
			v.setClipRect(
				((cnum & 1) and (gameWidth/2)*FRACUNIT or 0),
				((cnum > 1) and (gameHeight/2)*FRACUNIT or 0),
				(gameWidth/2)*FRACUNIT,
				(gameHeight/2)*FRACUNIT,
				flags
			)
		end
	end
	
	if cm.finishstart ~= -1 then -- Player has finished: move the claw back up!
		local t = leveltime - cm.finishstart
		local et = min(t, clawduration) * FRACUNIT / clawduration
		local y = clawy_f + ease.insine(et, 0, -64*FRACUNIT)
		
		v.drawScaled(clawx_f, y, FRACUNIT, clawopen, flags)
		v.drawScaled(clawx_f-((clawiconhwidth*FRACUNIT) + stupidoffset), y + iconpos*FRACUNIT, FRACUNIT, clawicon, flags, clawiconcolor)
	elseif leveltime >= (starttime-TICRATE) then -- Player is in race: claw is stagnant
		-- Normal drawing here...
		v.draw(clawx, clawy, clawclose, flags)
		-- Yes the offset is very important
		v.drawScaled(clawx_f - ((clawiconhwidth*FRACUNIT) + stupidoffset), clawy_f + iconpos*FRACUNIT, FRACUNIT, clawicon, flags, clawiconcolor)
		v.drawString(clawx, clawy + iconpos + clawiconhwidth*2, clawplayer, flags | V_TRANSLUCENT, "thin-center")
	elseif cm.clawstart ~= -1 then -- Player in position: move the claw down...
		local t = leveltime - cm.clawstart
		local et = min(t, clawduration) * FRACUNIT / clawduration
		local y = clawy_f + ease.outsine(et, -64*FRACUNIT, 0)
		
		v.drawScaled(clawx_f, y, FRACUNIT, clawopen, flags)
		if partner ~= nil then
			v.drawScaled(clawx_f-((clawiconhwidth*FRACUNIT) + stupidoffset), y + (iconpos+1)*FRACUNIT, FRACUNIT, clawicon, flags, clawiconcolor)
		else
			-- uhhh...
			local idx = ((leveltime / 3) % table.maxn(combinator.playerList)) + 1
			local mmap = v.getSprite2Patch(combinator.playerList[idx].mo.skin, SPR2_XTRA, false, clawiconidx)
			v.drawScaled(clawx_f-((clawiconhwidth*FRACUNIT) + stupidoffset), y + (iconpos+1)*FRACUNIT, FRACUNIT, mmap, flags, v.getColormap(combinator.playerList[idx].mo.skin, combinator.playerList[idx].mo.color))
		end
	end
	
	-- Starting line text! Because why not
	local text = cm.textmsg
	local sublen = 0
	if cm.textmsgstate == 0 then
		sublen = cm.textmsgtimer
	else
		sublen = string.len(text)
	end
	local subtext = string.sub(text, 0, sublen)
	
	if r_splitscreen < 2 and leveltime < starttime+(TICRATE*3) and leveltime > introtime and sublen > 0 then
		-- i actually like this custom drawstring thign
		WTXT.writeText(v, {
			x = clawx, y = clawy + 10,
			font = FONT_FANCYMENU,
			text = subtext,
			halign = 0,
			colormap = v.getStringColormap(cm.textmsgcolor),
			dflags = flags,
			wflags = DF_INTPOS,
			func = function(dd, b)
				local w = (b.autowidth/FRACUNIT)
				-- yuuuuup
				K_DrawSticker(v, clawx - w/2, clawy + 11, w + 2, (flags & ~V_30TRANS), true)
			end,
			predrawfunc = function(dd, b, l, p)
				p.yoff = V_DanceYOffset(p.num) * FRACUNIT/8
			end
		})
	end
	
	if v.interpolate ~= nil then v.interpolate(false) end
	
	if v.clearClipRect ~= nil then v.clearClipRect() end
end

local turngfx = {
	[0] = "",
	[1] = "CB_TLFT",
	[2] = "CB_TRHT",
	[3] = "CB_TSHT",
	[4] = "CB_TWWY",
}

local function turnsignal(v, p, c)
	if not (combinator.active) then return end
	if p.spectator then return end
	-- if splitscreen then return end
	
	if not (p.mo and p.mo.valid) then return end
	local mo = p.mo
	
	local partner, gargoyle = combinator.returnPartnerMobj(p, true)
	local cm = p.combiVars
	if not cm then return end
	
	if gargoyle then return end -- miserable
	
	-- get splitscreen state
	local r_splitscreen = SG_GetSplitscreen()
	local cnum = c.pnum - 1
	
	-- so that i dont have to edit each one by one
	local flags = V_SNAPTOLEFT|V_SNAPTOTOP
	
	-- give it some hud transparency
	if r_splitscreen == 1 then
		flags = $ | V_30TRANS
	end
	
	-- grab true screen width
	-- snap flags are a fuck
	-- i do not want to see them again
	-- fuck splitscreen too
	local screenWidth, screenHeight = v.width(), v.height()
	local screenScaleX, _frac = v.dupx()
	local screenScaleY, _frac = v.dupy()
	local gameWidth, gameHeight = screenWidth / screenScaleX, screenHeight / screenScaleY
	
	--
	-- RECEIVING TURN SIGNALS
	--
	
	local x, y = gameWidth / 2, gameHeight / 2
	
	-- x coords
	if (r_splitscreen > 1) then
		if cnum == 0 or cnum == 2 then
			x = $ - (gameWidth/4)
		else
			x = $ + (gameWidth/4)
		end
	end
	-- y coords
	if (r_splitscreen == 1 and (cnum == 1)) or (r_splitscreen > 1 and (cnum > 1)) then
		y = $ + (gameHeight/4)
	elseif (r_splitscreen == 1 and (cnum == 0)) or (r_splitscreen > 1 and (cnum < 2)) then
		y = $ - (gameHeight/4)
	end
	
	local turnsigsize = ((splitscreen > 1) and "S" or "B")
	local turnsigngfx = v.cachePatch(turngfx[cm.turnsignal_got]..turnsigsize)
	local turnhwidth = turnsigngfx.width/2
	local turnhheight = turnsigngfx.height/2
	
	if cm.turnsignal_cooldown > 0 and cm.turnsignal_got then
		local signal = cm.turnsignal_got
		local trans = abs(cm.turnsignal_cooldown / 5) % 3 + 6
		trans = $ << FF_TRANSSHIFT
		local offset = ((signal < 3) and ((signal == 1) and -turnhwidth*2 or turnhwidth*2) or 0)
		v.draw(x + offset, y - turnhheight, turnsigngfx, (flags & ~V_30TRANS)|trans)
	end
	
	--
	-- SENDING TURN SIGNALS
	--
	
	local x, y = gameWidth / 2, 0
	
	-- x coords
	if (r_splitscreen > 1) then
		if cnum == 0 or cnum == 2 then
			x = $ - (gameWidth/4)
		else
			x = $ + (gameWidth/4)
		end
	end
	-- y coords, just move it down...
	if (r_splitscreen == 1 and (cnum == 1)) or (r_splitscreen > 1 and (cnum > 1)) then
		y = $ + (gameHeight/2)
	end
	
	turnsigngfx = v.cachePatch(turngfx[cm.turnsignal_sent].."S")
	turnhwidth = turnsigngfx.width/2
	
	if cm.turnsignal_cooldown > 0 and cm.turnsignal_sent then
		v.draw(x, y + 48, turnsigngfx, flags | V_50TRANS)
	end
end

local function teleportspoofflashbang(v, p, c)
	if not (combinator.active) then return end
	if not (combinator.flashbang.value) then return end
	if p.spectator then return end
	-- if splitscreen then return end
	
	if not (p.mo and p.mo.valid) then return end
	local mo = p.mo
	
	local partner, gargoyle = combinator.returnPartnerMobj(p, true)
	local cm = p.combiVars
	if not cm then return end
	
	-- get splitscreen state
	local r_splitscreen = SG_GetSplitscreen()
	local cnum = c.pnum - 1
	
	-- so that i dont have to edit each one by one
	local flags = V_SNAPTOLEFT|V_SNAPTOTOP
	
	-- grab true screen width
	-- snap flags are a fuck
	-- i do not want to see them again
	-- fuck splitscreen too
	local screenWidth, screenHeight = v.width(), v.height()
	local screenScaleX, _frac = v.dupx()
	local screenScaleY, _frac = v.dupy()
	local gameWidth, gameHeight = screenWidth / screenScaleX, screenHeight / screenScaleY
	
	-- do a funny clip rect!
	if v.setClipRect ~= nil then
		if r_splitscreen == 1 then
			v.setClipRect(
				0, 
				((cnum == 1) and (gameHeight/2)*FRACUNIT or 0),
				gameWidth*FRACUNIT,
				(gameHeight/2)*FRACUNIT,
				V_SNAPTOLEFT|V_SNAPTOTOP
			)
		elseif r_splitscreen > 1 then
			v.setClipRect(
				((cnum & 1) and (gameWidth/2)*FRACUNIT or 0),
				((cnum > 1) and (gameHeight/2)*FRACUNIT or 0),
				(gameWidth/2)*FRACUNIT,
				(gameHeight/2)*FRACUNIT,
				V_SNAPTOLEFT|V_SNAPTOTOP
			)
		end
	end
	
	v.fadeScreen(0, min(10, cm.flashbangtics))
end

hud.add(partnertag, "game")
hud.add(teleportspoofflashbang, "game")
hud.add(newhudinfostuff, "game")
hud.add(turnsignal, "game")
-- cache the font immediately
local function fcache(v, p, c)
	if not cachedfonts then WTXT.fontSentinel(v) cachedfonts = true end
end
-- hud.add(fcache, "titlecard") -- may make titlecard delayed
hud.add(fcache, "intermission")