-- https://stackoverflow.com/a/16077650
-- I'm lazy.
local function deepcopy(o, seen)
	seen = seen or {}
	if o == nil then return nil end
	if seen[o] then return seen[o] end

	local no
	if type(o) == 'table' then
		no = {}
		seen[o] = no

		for k, v in next, o, nil do
			no[deepcopy(k, seen)] = deepcopy(v, seen)
		end
		setmetatable(no, deepcopy(getmetatable(o), seen))
	else -- number, string, boolean, etc
		no = o
	end
	return no
end

local function P_IsDisplayPlayer(p)
	if not (p and p.valid) then return false end
	
	-- There are only 4 display players at a time... right?
	for i = 0, 3 do
		if not (displayplayers[i] and displayplayers[i].valid) then continue end
		if p == displayplayers[i] then return true end
	end
	
	return false
end

-- Creates the persistent combi vars table.
-- This cannot be forced to be empty.
function combinator.initCombiPersVars(player)
	if player.combiPersVars then return end
	player.combiPersVars = {}
	local cm = player.combiPersVars
	
	-- FRIENDS STUFF (SHOULD NOT BE RESET.)
	cm.friendpending = -1 -- Pending friend. -1 denotes having none.
	cm.friend_num = -1 -- Target player number to use as a friend. -1 denotes having none.
end

-- Creates the combi vars table.
function combinator.initCombiVars(player, force)
	if player.combiVars and not force then return end
	
	player.combiVars = {} -- init this shit
	
	local cm = player.combiVars
	cm.partner = nil -- your partner player
	cm.gargoyle = nil -- gargoyle mobj, if applicable.
	
	cm.pull = {xy = 0, z = 0} -- pull force.
	-- rank stuff
	cm.team_rank = nil
	cm.old_team_rank = nil
	cm.team_rank_cooldown = 0
	cm.real_rank = 0
	-- for loops
	cm.markedforloop = 0
	
	cm.jawztarget = nil -- jawz target, without your partner being targetted
	cm.fuckedOver = false -- partner assigned at the end?
	
	-- HUD STUFFS
	cm.clawstart = -1 -- claw start time, for easing.
	cm.finishstart = -1 -- finish leveltime, for easing.
	cm.fuckedOverTime = 0 -- fucked over, time to flash
	-- text messages, similar to K_AddMessage
	cm.textmsg = ""
	cm.textmsgstate = 0
	cm.textmsgtimer = -1
	cm.textmsgcolor = 0
	-- Turn signals
	cm.turnsignal_sent = 0
	cm.turnsignal_got = 0
	cm.turnsignal_cooldown = 0
	-- Teleport spoof flashbang
	cm.flashbangtics = 0
	
	-- a copy of our combi vars, for gargoyle related things.
	-- ALWAYS DO THIS LAST
	-- (well except for friends...)
	cm.gargoyleVars = deepcopy(cm)
end

-- Returns the partner's mobj and a boolean if its a gargoyle or not.
-- Also summons the gargoyle if it doesn't exist.
function combinator.returnPartnerMobj(player, skipgargoyle)
	local cm = player.combiVars
	if not cm then return nil, false end -- you...
	if cm.partner and cm.partner.valid and not cm.partner.spectator then
		return cm.partner.mo, false
	end
	if leveltime <= starttime-TICRATE then return nil, true end
	if skipgargoyle then return cm.gargoyle, true end
	-- gargoyle, return dummy
	if not (cm.gargoyle and cm.gargoyle.valid) then
		cm.gargoyle = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + 192*mapobjectscale, MT_GARGOYLE)
		cm.gargoyle.target = player.mo
		-- force refresh
		combinator.drawCombiRing(player.mo, cm.gargoyle, true, true)
	end
	return cm.gargoyle, true
end

-- Handles pulling variables.
local function combipull(player)
	local partner, gargoyle = combinator.returnPartnerMobj(player)
	local mo = player.mo
	local cm = player.combiVars
	local pcm = ((partner and partner.player) and partner.player.combiVars or cm.gargoyleVars)
	local cmd = player.cmd
	
	-- reset
	cm.pull.xy, pcm.pull.xy = mapobjectscale/2, mapobjectscale/2
	cm.pull.z, pcm.pull.z = mapobjectscale/2, mapobjectscale/2
	
	-- pull at half the amount
	if cmd.buttons & BT_ACCELERATE then
		cm.pull.xy = $ + mapobjectscale/2
	end
	-- pull at the full amount
	if cmd.buttons & BT_BRAKE then
		cm.pull.xy = $ + mapobjectscale
	end
	
	-- set pull z
	if not P_IsObjectOnGround(mo) then
		if gargoyle then -- dont bother
			pcm.pull.xy = 0
			pcm.pull.z = 0
		else
			cm.pull.z = $ + mapobjectscale*2
		end
	else
		cm.pull.z = 0
	end
	
	-- might as well update it here
	if not gargoyle then
		-- pull at half the amount
		cmd = partner.player.cmd
		if cmd.buttons & BT_ACCELERATE then
			pcm.pull.xy = $ + mapobjectscale/2
		end
		-- pull at the full amount
		if cmd.buttons & BT_BRAKE then
			pcm.pull.xy = $ + mapobjectscale
		end
		
		-- set pull z
		if not P_IsObjectOnGround(partner) then
			pcm.pull.z = $ + mapobjectscale*2
		else
			pcm.pull.z = 0
		end
	end
end

-- can someone at kartkrew expose P_SetupSignObject this is exhausting
local function P_SetupSignObject(mo)
	local sign = P_SpawnMobj(mo.x, mo.y, mo.floorz, MT_SIGN)
	-- Yes, we need to snap. Otherwise, the sign would spin indefinitely.
	-- future haya: fuck it.
	sign.angle = mo.angle + ANGLE_180
		
	-- auuuugh
	local cur = sign
	local prev = nil
		
	-- set up the sign itself
	sign.scale = 5*($/2)
	sign.target = mo
	sign.state = S_SIGN_POLE
		
	sign.movefactor = sign.z
	P_SetOrigin(sign, sign.x, sign.y, sign.z + ((576*sign.scale) * P_MobjFlip(sign)))
	sign.movecount = 1
	sign.extravalue1 = AngleFixed(sign.angle) >> FRACBITS
		
	-- setup the overlay pieces
	-- front
	cur.hnext = P_SpawnMobjFromMobj(sign, 0, 0, 0, MT_SIGN_PIECE)
	cur.hnext.target = sign
	cur.hnext.state = S_SIGN_FACE
	cur.hnext.extravalue1 = 6
	cur.hnext.extravalue2 = 0
		
	prev = cur
	cur = cur.hnext
	cur.hprev = prev
		
	-- player icon
	cur.hnext = P_SpawnMobjFromMobj(sign, 0, 0, 0, MT_SIGN_PIECE)
	cur.hnext.target = sign
	cur.hnext.skin = mo.skin
	cur.hnext.state = S_KART_SIGN
	cur.hnext.extravalue1 = 7
	cur.hnext.extravalue2 = 0
		
	prev = cur
	cur = cur.hnext
	cur.hprev = prev
		
	-- back
	cur.hnext = P_SpawnMobjFromMobj(sign, 0, 0, 0, MT_SIGN_PIECE)
	cur.hnext.target = sign
	cur.hnext.state = S_SIGN_BACK
	cur.hnext.extravalue1 = 6
	cur.hnext.extravalue2 = 2
		
	prev = cur
	cur = cur.hnext
	cur.hprev = prev
		
	-- sides
	cur.hnext = P_SpawnMobjFromMobj(sign, 0, 0, 0, MT_SIGN_PIECE)
	cur.hnext.target = sign
	cur.hnext.state = S_SIGN_SIDE
	cur.hnext.extravalue1 = 30
	cur.hnext.extravalue2 = 1
		
	prev = cur
	cur = cur.hnext
	cur.hprev = prev
		
	cur.hnext = P_SpawnMobjFromMobj(sign, 0, 0, 0, MT_SIGN_PIECE)
	cur.hnext.target = sign
	cur.hnext.state = S_SIGN_SIDE
	cur.hnext.extravalue1 = 30
	cur.hnext.extravalue2 = 3
		
	prev = cur
	cur = cur.hnext
	cur.hprev = prev
end

-- Handle syncing of timers between partners.
local function combisync(player)
	local partner, gargoyle = combinator.returnPartnerMobj(player)
	local mo = player.mo
	local cm = player.combiVars
	local pcm = ((partner and partner.player) and partner.player.combiVars or cm.gargoyleVars)
	
	if player.exiting and cm.finishstart == -1 then
		if P_IsDisplayPlayer(player) then
			S_StartSound(nil, sfx_kc37)
		end
		cm.finishstart = leveltime
		if not gargoyle then 
			pcm.finishstart = leveltime 
			if P_IsDisplayPlayer(partner.player) then
				S_StartSound(nil, sfx_kc37)
			end
		end
	end
	
	if gargoyle then return end -- Who are we fucking syncing to?
	
	local pp = partner.player
	
	if player.growshrinktimer < 0 then -- shrink
		pp.growshrinktimer = min($, player.growshrinktimer) -- always go for the smallest
	elseif player.growshrinktimer > 0 then -- grow
		pp.growshrinktimer = max($, player.growshrinktimer)
		-- update scale
		partner.destscale = max($, mo.destscale)
		-- partner.destspeed = max($, mo.destspeed)
	end
	
	-- update various timers
	pp.sneakertimer = max($, player.sneakertimer)
	pp.invincibilitytimer = max($, player.invincibilitytimer)
	pp.hyudorotimer = max($, player.hyudorotimer)
	pp.driftboost = max($, player.driftboost)
	pp.strongdriftboost = max($, player.strongdriftboost)
	-- pp.startboost = max($, player.startboost) -- How the fuck
	pp.ringboost = max($, player.ringboost)
	pp.spindashboost = max($, player.spindashboost)
	pp.wavedashboost = max($, player.wavedashboost)
	pp.gateboost = max($, player.gateboost)
	pp.gatesound = max($, player.gatesound)
	pp.tiregrease = max($, player.tiregrease)
	-- :thinking:
	if pp.superring == 0 and player.superring ~= 0 then
		-- making this a max function would be too much
		pp.superring = player.superring
	end
	-- for race
	pp.realtime = max($, player.realtime)
	-- for bottl
	pp.roundscore = max($, player.roundscore)
	pp.spheres = max($, player.spheres)
	pp.emeralds = max($, player.emeralds)
	pp.spheredigestion = max($, player.spheredigestion)
	
	
	-- no lap updating for now...
	
	-- handle player exit...
	if player.position == 1 and player.exiting and not pp.exiting then
		P_SetupSignObject(partner)
		-- sign is setup, lets actually make the player exit!
		P_DoPlayerExit(pp, 0)
	end
end

-- Handles turn signals.
local function combisignals(player)
	local partner, gargoyle = combinator.returnPartnerMobj(player)
	local mo = player.mo
	local cm = player.combiVars
	local pcm = ((partner and partner.player) and partner.player.combiVars or cm.gargoyleVars)

	-- HUD: Handle turn signals
	cm.turnsignal_cooldown = max(0, $ - 1)
	if cm.turnsignal_cooldown == 0 then -- no cooldown, so attempt to do turn signal shit
		local buttons = player.cmd.buttons
		local turntype = 0
		if (player.cmd.buttons & BT_LUAA) then -- left
			turntype = 1
			--print("sent turn signal: 1")
			cm.turnsignal_cooldown = TICRATE
			pcm.turnsignal_cooldown = TICRATE
		end
		if (player.cmd.buttons & BT_LUAB) then -- right
			turntype = 2
			--print("sent turn signal: 2")
			cm.turnsignal_cooldown = TICRATE
			pcm.turnsignal_cooldown = TICRATE
		end
		if (player.cmd.buttons & BT_LUAC) then -- straight
			turntype = 3
			--print("sent turn signal: 3")
			cm.turnsignal_cooldown = TICRATE
			pcm.turnsignal_cooldown = TICRATE
		end
		if (player.cmd.buttons & BT_LUAC) and (player.cmd.buttons & BT_ACCELERATE) and (player.cmd.buttons & BT_BRAKE) then -- 'come back! try again!'
			turntype = 4
			--print("sent turn signal: 4")
			cm.turnsignal_cooldown = TICRATE
			pcm.turnsignal_cooldown = TICRATE
		end
		cm.turnsignal_sent = turntype
		pcm.turnsignal_got = turntype
	end
end

-- Handles respawn logic.
local function combirespawn(player)
	local partner, gargoyle = combinator.returnPartnerMobj(player)
	local mo = player.mo
	local cm = player.combiVars
	local pcm = ((partner and partner.player) and partner.player.combiVars or cm.gargoyleVars)

	-- we are moving....
	if player.respawn.state == 1 then
		if gargoyle and partner then -- yknow what? just move the gargoyle with you.
			P_MoveOrigin(partner, mo.x, mo.y, mo.z)
			partner.flags = $ | MF_NOCLIP -- noclip on it for now
			return
		end
		if not partner then return end
		-- We have a partner!
		local pp = partner.player
		pp.respawn.state = 1
		-- Node priorities, sorry!
		pp.respawn.pointx = player.respawn.pointx
		pp.respawn.pointy = player.respawn.pointy
		pp.respawn.pointz = player.respawn.pointz
		-- Use the smallest timer! Stalling is the worst thing here
		pp.respawn.timer = min($, player.respawn.timer)
	elseif player.respawn.state == 0 then
		if gargoyle and partner then
			partner.flags = $ & ~MF_NOCLIP -- remove partner gargoyle noclip
		end
	end
end

local clawduration = TICRATE

local function C_AddNewMessage(cm, msg)
	cm.textmsg = msg
	cm.textmsgstate = 0
	cm.textmsgtimer = 0
end

-- Handles stuff related to visuals n such. Should be called in ThinkFrame.
local function combimisc(player)
	local partner, gargoyle = combinator.returnPartnerMobj(player)
	local mo = player.mo
	local cm = player.combiVars
	local pcm = ((partner and partner.player) and partner.player.combiVars or cm.gargoyleVars)
	
	-- HUD: Force it down!
	if leveltime == starttime-clawduration-(TICRATE*2) and cm.clawstart == -1 then
		cm.clawstart = leveltime
		if P_IsDisplayPlayer(player) then
			S_StartSound(nil, sfx_kc37)
		end
	elseif leveltime == starttime-TICRATE then -- confirm!
		if P_IsDisplayPlayer(player) then
			S_StartSound(nil, sfx_kc68)
		end
	end
	
	-- HUD: Silence.
	if cm.clawstart ~= -1 and leveltime == cm.clawstart + clawduration and P_IsDisplayPlayer(player) then
		S_StopSoundByID(nil, sfx_kc37)
	end
	if cm.finishstart ~= -1 and leveltime == cm.finishstart + clawduration and P_IsDisplayPlayer(player) then
		S_StopSoundByID(nil, sfx_kc37)
	end
	
	-- HUD: Increment timer for messages!
	if cm.textmsg ~= "" and cm.textmsgstate == 0 then
		cm.textmsgtimer = $ + 1
		if cm.textmsgtimer > string.len(cm.textmsg) then
			cm.textmsgstate = 1
		end
	end
	
	-- HUD: Timed messages!
	if leveltime == introtime then
		if P_IsDisplayPlayer(player) then
			S_StartSound(nil, sfx_s3k47)
		end
		C_AddNewMessage(cm, "Find a partner!")
	end
	
	-- HUD: Clear out this shit
	if leveltime >= starttime+3*TICRATE then
		cm.textmsg = ""
		cm.textmsgstate = 0
		cm.textmsgtimer = 0
		cm.textmsgcolor = 0
	end
	
	-- HUD: Update fucked over flashing.
	cm.fuckedOverTime = max(0, $-1)
	cm.flashbangtics = max(0, $-1)
	
	if not partner then return end
	
	-- move combi ring
	if partner then
		combinator.drawCombiRing(mo, partner, false, gargoyle)
	end
end

-- Handles the final distance yanking. ALWAYS CALL THIS LAST.
local function combiyank(player)
	local partner, gargoyle = combinator.returnPartnerMobj(player)
	local mo = player.mo
	local cm = player.combiVars
	local pcm = ((partner and partner.player) and partner.player.combiVars or cm.gargoyleVars)
	
	if not partner then return end
	
	-- Originally divided by 4, which was too much! Dial it down!
	local yank_xy = FixedDiv(pcm.pull.xy, pcm.pull.xy + cm.pull.xy) / 12
	local yank_z = 0
	if pcm.pull.z ~= 0 and cm.pull.z ~= 0 then
		yank_z = FixedDiv(pcm.pull.z, pcm.pull.z + cm.pull.z) / 12
	end
	
	local distance = FixedHypot(FixedHypot(mo.x - partner.x, mo.y - partner.y), mo.z - partner.z)
	
	cm.markedforloop = max(0, $-1)
	
	cm.lastX = $ or 0
	cm.lastY = $ or 0
	cm.lastZ = $ or 0
	pcm.lastX = $ or 0
	pcm.lastY = $ or 0
	pcm.lastZ = $ or 0
	
	-- Teleport spoof
	if combinator.teleportspoof.value > 0 and distance > combinator.teleportspoof.value*mapobjectscale and player.respawn.state == 0 then 
		-- For this, take priority.
		-- Which partner has the biggest change in coordinates from the last frame
		-- If it was the partner, teleport us
		-- If it was the player, teleport the partner
		--
		-- TODO: For now, no Z coordinate checking. Had too many problems especially at death egg
		local pdist = FixedHypot(mo.x - cm.lastX, mo.y - cm.lastY)
		local cdist = FixedHypot(partner.x - pcm.lastX, partner.y - pcm.lastY)
		
		local offending, teleport = nil, nil
		
		if pdist > cdist or gargoyle then -- gargoyles should just teleport to you man
			offending = mo
			teleport = partner
			-- print("PLAYER HAS PRIORITY")
		else
			offending = partner
			teleport = mo
			-- print("PARTNER HAS PRIORITY")
		end
		
		-- For some reason this doesnt work...
		-- future haya: player->flashing isnt exposed to be edited :melting_face:
		
-- 		if offending.player then offending.player.flashing = 20 end
-- 		if teleport.player then teleport.player.flashing = 20 end
		
		P_SetOrigin(teleport, offending.x, offending.y, offending.z)
		if offending.eflags & MFE_VERTICALFLIP then
			teleport.eflags = $ | MFE_VERTICALFLIP
		else
			teleport.eflags = $ & ~MFE_VERTICALFLIP
		end
		if teleport.player then
			local pp = teleport.player
			pp.respawn.state = 1
			pp.respawn.pointx = offending.x
			pp.respawn.pointy = offending.y
			pp.respawn.pointz = offending.z
			pp.respawn.timer = 5
			pp.combiVars.flashbangtics = 20
			
			if P_IsDisplayPlayer(pp) then
				S_StartSoundAtVolume(nil, sfx_mixup, 224)
			end
		end
		teleport.angle = offending.angle
		return
	end
	
	cm.lastX = mo.x
	cm.lastY = mo.y
	cm.lastZ = mo.z
	pcm.lastX = partner.x
	pcm.lastY = partner.y
	pcm.lastZ = partner.z
	
	-- dont yank if we dont have enough distance away
	-- Originally 450. More speed in this game, so give it some lineage!
	if distance < 560*mapobjectscale then return end
	
	-- On loops, if one of you has this flag, do not do distance checks. Please.
	-- Yes this is the only way to check for loops, dammit
	if (mo.flags & MF_NOCLIPHEIGHT) or (partner.flags & MF_NOCLIPHEIGHT) then
		cm.markedforloop = 10
		pcm.markedforloop = 10
		return
	end
	
	-- marked for loop. do not do distance checks until we are close again
	-- this might cause exploits...
	if cm.markedforloop then return end
	
	distance = $ - (560*mapobjectscale)
	
	local angle = R_PointToAngle2(mo.x, mo.y, partner.x, partner.y)
	local v_angle = R_PointToAngle2(0, mo.z, R_PointToDist2(mo.x, mo.y, partner.x, partner.y), partner.z)
	
	-- do the yank stuff
	P_Thrust(mo, angle, FixedMul(FixedMul(cos(v_angle), yank_xy), distance))
	mo.momz = $ + FixedMul(FixedMul(sin(v_angle), yank_z), distance) -- TODO: account for mobj flip
	
	if gargoyle then -- manually move the gargoyle!
		P_Thrust(partner, angle, -FixedMul(FixedMul(cos(v_angle), mapobjectscale-yank_xy), distance))
		partner.momz = $ - FixedMul(FixedMul(sin(v_angle), mapobjectscale-yank_z), distance)
	end
end

local col = {
	[MT_ORBINAUT] = true,
	[MT_ORBINAUT_SHIELD] = true,
	[MT_JAWZ] = true,
	[MT_JAWZ_SHIELD] = true,
	[MT_BALLHOG] = true,
	[MT_BANANA] = true,
	[MT_BANANA_SHIELD] = true,
	[MT_EGGMANITEM] = true,
	[MT_EGGMANITEM_SHIELD] = true,
	[MT_DROPTARGET] = true,
	[MT_DROPTARGET_SHIELD] = true,
}

-- Handles collision with partner hazards.
local function DetectPartnerCollision(thing, thing2)
	if not (combinator.active) then return end
	if combinator.friendlyfire.value then return end -- friendly fire on
	local cm = thing.player.combiVars
	if not cm then return end
	if col[thing2.type] then
		local targ = thing2.target -- target is the colliding thing's owner here
		if targ.player and targ.player == cm.partner then
			return false
		end
	end
	return
end

-- Handles general partner damage.
addHook("ShouldDamage", function(mo, inflictor, source, damage)
	if not (combinator.active) then return end
	if combinator.friendlyfire.value then return end -- friendly fire on
    if not (mo and mo.valid and mo.player) then return end
	local mop = mo.player
	if not mop.combiVars then return end
    if damage < 10000 then
		if inflictor 
		and inflictor.valid 
		and inflictor.player
		and inflictor.player == mop.combiVars.partner then
			return false
		end
		if source 
		and source.valid 
		and source.player
		and source.player == mop.combiVars.partner then
			return false
		end
	end
end)

addHook("MobjCollide", DetectPartnerCollision, MT_PLAYER)
addHook("MobjMoveCollide", DetectPartnerCollision, MT_PLAYER)

-- Handles partner grabbing! First come, first serve!
local function combigetpartner(thing, thing2)
	if not (combinator.active) then return end
	if thing2.type ~= MT_PLAYER then return end
	
	local pone, ptwo = thing.player, thing2.player
	if not (pone.combiVars and ptwo.combiVars) then return end
	
	local cm, pcm = pone.combiVars, ptwo.combiVars
	if leveltime >= starttime-TICRATE then return end
	if not combinator.positiontag.value and (cm.partner or pcm.partner) then return end
	
	S_StopSoundByID(thing, sfx_kc37)
	S_StopSoundByID(thing2, sfx_kc37)
	
	if P_IsDisplayPlayer(pone) then S_StartSound(nil, sfx_kc37) end
	if P_IsDisplayPlayer(ptwo) then S_StartSound(nil, sfx_kc37) end
	
	C_AddNewMessage(cm, "Please wait...")
	cm.textmsgcolor = V_GREENMAP
	C_AddNewMessage(pcm, "Please wait...")
	pcm.textmsgcolor = V_GREENMAP
	
	if cm.partner and cm.partner ~= ptwo then -- They have a partner? Get rid of them.
		cm.partner.combiVars.partner = nil
		cm.partner = nil
	end
	if pcm.partner and pcm.partner ~= pone then -- Same with you, idiot.
		pcm.partner.combiVars.partner = nil
		pcm.partner = nil
	end
	
	cm.partner = ptwo
	pcm.partner = pone
	
	cm.clawstart = leveltime
	pcm.clawstart = leveltime
end

addHook("MobjCollide", combigetpartner, MT_PLAYER)
addHook("MobjMoveCollide", combigetpartner, MT_PLAYER)

-- prevents shooting your own teammate
local function CombiJawzRehandler(player)
	local partner, gargoyle = combinator.returnPartnerMobj(player)
	local mo = player.mo
	local cm = player.combiVars
	local pcm = ((partner and partner.player) and partner.player.combiVars or cm.gargoyleVars)
	
	if gargoyle then return end -- No need, dumbass.
	
	cm.jawztarget = $ or player.lastjawztarget
	-- TODO: Figure out a way if you're actually holding the item out.
	if player.itemtype == KITEM_JAWZ and cm.jawztarget and cm.jawztarget ~= -1 then
		local lasttarg = cm.jawztarget
		local targ = nil
		
		if (player.jawztargetdelay and not (players[lasttarg].spectator or (players[lasttarg].combiVars.partner == player or cm.partner == players[lasttarg]))) then
			targ = players[lasttarg]
			player.jawztargetdelay = $ - 1
		else
			targ = K_FindJawzTarget(player.mo, player)
		end
			
		if (not targ or not targ.mo) then
			cm.jawztarget = -1
			player.jawztargetdelay = 0
			return
		end
		
		if (#targ != lasttarg) then
			if P_IsDisplayPlayer(player) or P_IsDisplayPlayer(targ) then
				S_StopSoundByID(nil, sfx_s3k89)
				S_StartSound(nil, sfx_s3k89)
			else
				S_StopSoundByID(targ.mo, sfx_s3k89)
				S_StartSound(targ.mo, sfx_s3k89)
			end

			cm.jawztarget = #targ
			player.jawztargetdelay = 5
		end
	else
		cm.jawztarget = -1
		player.jawztargetdelay = 0
	end
	
	player.lastjawztarget = cm.jawztarget
end

-- yeah
addHook("MapLoad", function()
	combinator.active = combinator.enabled.value
	combinator.partnerSelected = false
	if not combinator.active then return end
	-- Init all combi vars
	for p in players.iterate do
		if not (p.mo and p.mo.valid) or p.spectator then continue end -- mo doesnt exist or is a spectator
		combinator.initCombiVars(p, true)
		combinator.initCombiPersVars(p)
	end
end)
addHook("PlayerSpawn", function(p) -- Look, just.
	if not combinator.active then return end
	combinator.initCombiVars(p)
	combinator.initCombiPersVars(p)
end)
addHook("NetVars", function(net) combinator.active = net($) combinator.partnerSelected = net($) end)

-- Roulette thing
addHook("ThinkFrame", function()
	if not combinator.active then return end
	
	-- for hud
	combinator.playerList = {}
	
	for p in players.iterate do
		if not (p.mo and p.mo.valid) and p.spectator then 
			p.combiVars = nil 
			continue
		end
		combimisc(p)
		table.insert(combinator.playerList, p)
	end
	
	if leveltime == starttime-TICRATE then
		-- Add all in-game players to a list
		local plist = {}
		for p in players.iterate do
			if not (p.mo and p.mo.valid) or p.spectator then continue end -- mo doesnt exist or is a spectator
			C_AddNewMessage(p.combiVars, "Good luck!")
			p.combiVars.textmsgcolor = V_SKYMAP
			if p.combiVars and p.combiVars.partner then continue end -- we already have one!
			-- Friend is here, speed off!
			combinator.initCombiPersVars(p)
			local cm = p.combiPersVars
			if cm.friend_num > -1 and players[cm.friend_num] then
				local fren = players[cm.friend_num]
				if not (fren.mo and fren.mo.valid) or fren.spectator then table.insert(plist, p) continue end -- dont bother
				if fren.combiPersVars.friend_num ~= #p then table.insert(plist, p) continue end -- dont bother
				-- we have a friend, so pair them up!
				cm.partner = fren
				fren.combiVars.partner = p
				continue
			end
			-- Oh...
			p.combiVars.fuckedOver = true
			p.combiVars.fuckedOverTime = TICRATE
			C_AddNewMessage(p.combiVars, "The council decides your fate...")
			p.combiVars.textmsgcolor = V_REDMAP
			table.insert(plist, p)
		end
		
		-- While this player list size is bigger than 1...
		while #plist > 1 do
			-- Pick two random players...
			local one = table.remove(plist, P_RandomRange(1, #plist))
			local two = table.remove(plist, P_RandomRange(1, #plist))
			-- And assign each other to their partner
			one.combiVars.partner = two
			two.combiVars.partner = one
			
			S_StartSoundAtVolume(nil, sfx_slot04, 128, one)
			S_StartSoundAtVolume(nil, sfx_slot04, 128, two)
		end
		
		-- Ok, we still have someone...
		if #plist > 0 then
			-- Force summon a gargoyle. Your loss.
			local cm = plist[1].combiVars
			if not (cm.gargoyle and cm.gargoyle.valid) then
				cm.gargoyle = P_SpawnMobj(plist[1].mo.x, plist[1].mo.y, plist[1].mo.z + 192*mapobjectscale, MT_GARGOYLE)
				cm.gargoyle.target = plist[1].mo
				-- force refresh
				combinator.drawCombiRing(plist[1].mo, cm.gargoyle, true, true)
			end
			S_StartSoundAtVolume(nil, sfx_itrole, 255, plist[1])
		end
		
		combinator.partnerSelected = true
		
		return
	end
	
	-- if leveltime <= starttime-TICRATE then return end
	
	-- handle rankings
	local individual_rank = {}
	for p in players.iterate do
		if not (p.mo and p.mo.valid) or p.spectator then continue end -- mo doesnt exist or is a spectator
		local cm = p.combiVars
		if not cm then continue end -- how????
		
		cm.team_rank = nil
		cm.real_rank = $ or p.position
			
		local spot = 1
		while individual_rank[spot] and individual_rank[spot].combiVars.real_rank <= cm.real_rank do
			spot = $+1
		end

		table.insert(individual_rank, spot, p)
	end
	
	local rank = 1
	local team_rank = 0
	
	while individual_rank[rank] do
		local p = individual_rank[rank]
		p.combiVars.real_rank = nil
		-- That's a mouthful.
		if (p.combiVars.partner and p.combiVars.partner.valid) and p.combiVars.partner.combiVars.team_rank then
			p.combiVars.team_rank = p.combiVars.partner.combiVars.team_rank
		else
			team_rank = $ + 1
			p.combiVars.team_rank = team_rank
		end
		
		if p.combiVars.team_rank ~= p.combiVars.old_team_rank then
			p.combiVars.team_rank_cooldown = 10
			p.combiVars.old_team_rank = p.combiVars.team_rank
		elseif p.combiVars.team_rank_cooldown > 0 then
			p.combiVars.team_rank_cooldown = $ - 1
		end
		
		-- Update position stuff
		p.position = p.combiVars.team_rank
		p.positiondelay = p.combiVars.team_rank_cooldown
		
		rank = $ + 1
	end
end)

-- Force rankings here to make SPB distance checking work as intended
addHook("MobjThinker", do
	if not (combinator.active) then return end
	
	for p in players.iterate do
		if not p.combiVars.real_rank then
			p.combiVars.real_rank = p.position
		end

		if p.combiVars.team_rank then
			p.position = p.combiVars.team_rank
		end
	end
end, MT_PLAYER)

-- Handle combi logic
addHook("PlayerThink", function(p)
	if not (combinator.active) then return end
	if not (p.mo and p.mo.valid) or p.spectator then return end -- mo doesnt exist or is a spectator
	
	combipull(p)
	CombiJawzRehandler(p)
	combisync(p)
	combisignals(p)
	combirespawn(p)
	combiyank(p)
end)